@@ -7,12 +7,15 @@ Borrowed from Brad Chapman's implementation:
 https://github.com/chapmanb/bcbio-nextgen/blob/master/bcbio/distributed/ipython.py
 """
 import contextlib
+import math
 import os
 import pipes
 import uuid
 import shutil
 import subprocess
 import time
+from distutils.version import LooseVersion
+import sys
 
 from IPython.parallel import Client
 from IPython.parallel.apps import launcher
@@ -42,23 +45,30 @@ class BcbioLSFEngineSetLauncher(launcher.LSFEngineSetLauncher):
     """
     batch_file_name = Unicode(unicode("lsf_engine" + str(uuid.uuid4())))
     cores = traitlets.Integer(1, config=True)
+    mem = traitlets.Unicode("", config=True)
     default_template = traitlets.Unicode("""#!/bin/sh
 #BSUB -q {queue}
 #BSUB -J bcbio-ipengine[1-{n}]
 #BSUB -oo bcbio-ipengine.bsub.%%J
 #BSUB -n {cores}
 #BSUB -R "span[hosts=1]"
+{mem}
 %s %s --profile-dir="{profile_dir}" --cluster-id="{cluster_id}"
     """ % (' '.join(map(pipes.quote, launcher.ipengine_cmd_argv)),
            ' '.join(timeout_params)))
 
     def start(self, n):
         self.context["cores"] = self.cores
+        if self.mem:
+            # scale memory to kb
+            mem = int(float(self.mem) * 1024.0 * 1024.0)
+            self.context["mem"] = "#BSUB -M %s" % mem
+        else:
+            self.context["mem"] = ""
         return super(BcbioLSFEngineSetLauncher, self).start(n)
 
 class BcbioLSFControllerLauncher(launcher.LSFControllerLauncher):
     batch_file_name = Unicode(unicode("lsf_controller" + str(uuid.uuid4())))
-    "-".join( str(uuid.uuid4))
     default_template = traitlets.Unicode("""#!/bin/sh
 #BSUB -J bcbio-ipcontroller
 #BSUB -oo bcbio-ipcontroller.bsub.%%J
@@ -76,6 +86,7 @@ class BcbioSGEEngineSetLauncher(launcher.SGEEngineSetLauncher):
     cores = traitlets.Integer(1, config=True)
     pename = traitlets.Unicode("", config=True)
     resources = traitlets.Unicode("", config=True)
+    mem = traitlets.Unicode("", config=True)
     default_template = traitlets.Unicode("""#$ -V
 #$ -cwd
 #$ -b y
@@ -85,6 +96,7 @@ class BcbioSGEEngineSetLauncher(launcher.SGEEngineSetLauncher):
 #$ -N bcbio-ipengine
 #$ -t 1-{n}
 #$ -pe {pename} {cores}
+{mem}
 {resources}
 echo \($SGE_TASK_ID - 1\) \* 0.5 | bc | xargs sleep
 %s %s --profile-dir="{profile_dir}" --cluster-id="{cluster_id}"
@@ -93,6 +105,10 @@ echo \($SGE_TASK_ID - 1\) \* 0.5 | bc | xargs sleep
 
     def start(self, n):
         self.context["cores"] = self.cores
+        if self.mem:
+            self.context["mem"] = "#$ -l mem_free=%sG" % self.mem
+        else:
+            self.context["mem"] = ""
         self.context["pename"] = str(self.pename)
         self.context["resources"] = "\n".join(["#$ -l %s" % r.strip()
                                                for r in str(self.resources).split(";")
@@ -139,7 +155,7 @@ class SLURMLauncher(launcher.BatchSystemLauncher):
         help="The SLURM submit command ['sbatch']")
     # Send SIGKILL instead of term, otherwise the job is "CANCELLED", not
     # "FINISHED"
-    delete_command = List(['scancel --signal=KILL'], config=True,
+    delete_command = List(['scancel', '--signal=KILL'], config=True,
         help="The SLURM delete command ['scancel']")
     job_id_regexp = CRegExp(r'\d+', config=True,
         help="A regular expression used to get the job id from the output of 'sbatch'")
@@ -152,7 +168,48 @@ class SLURMLauncher(launcher.BatchSystemLauncher):
 
 
 class BcbioSLURMEngineSetLauncher(SLURMLauncher, launcher.BatchClusterAppMixin):
-    """Launch engines using SLURM"""
+    """Custom launcher handling heterogeneous clusters on SLURM
+    """
+    batch_file_name = Unicode(unicode("SLURM_engine" + str(uuid.uuid4())))
+    cores = traitlets.Integer(1, config=True)
+    mem = traitlets.Unicode("", config=True)
+    default_template = traitlets.Unicode("""#!/bin/sh
+#SBATCH -p {queue}
+#SBATCH -J bcbio-ipengine[1-{n}]
+#SBATCH -o bcbio-ipengine.out.%%j
+#SBATCH -e bcbio-ipengine.err.%%j
+#SBATCH --cpus-per-task={cores}
+#SBATCH --array=1-{n}
+{mem}
+%s %s --profile-dir="{profile_dir}" --cluster-id="{cluster_id}"
+    """ % (' '.join(map(pipes.quote, launcher.ipengine_cmd_argv)),
+           ' '.join(timeout_params)))
+
+    def start(self, n):
+        self.context["cores"] = self.cores
+        if self.mem:
+            # scale memory to Mb and divide by cores
+            mem = int(math.ceil(float(self.mem) * 1024.0 / self.cores))
+            self.context["mem"] = "#SBATCH --mem-per-cpu=%s" % mem
+        else:
+            self.context["mem"] = ""
+        return super(BcbioSLURMEngineSetLauncher, self).start(n)
+
+class BcbioSLURMControllerLauncher(SLURMLauncher, launcher.BatchClusterAppMixin):
+    batch_file_name = Unicode(unicode("SLURM_controller" + str(uuid.uuid4())))
+    default_template = traitlets.Unicode("""#!/bin/sh
+#SBATCH -J bcbio-ipcontroller
+#SBATCH -o bcbio-ipcontroller.out.%%j
+#SBATCH -e bcbio-ipcontroller.err.%%j
+%s --ip=* --log-to-file --profile-dir="{profile_dir}" --cluster-id="{cluster_id}" %s
+    """%(' '.join(map(pipes.quote, launcher.ipcontroller_cmd_argv)),
+         ' '.join(controller_params)))
+    def start(self):
+        return super(BcbioSLURMControllerLauncher, self).start(1)
+
+
+class BcbioOLDSLURMEngineSetLauncher(SLURMLauncher, launcher.BatchClusterAppMixin):
+    """Launch engines using SLURM for version < 2.6"""
     machines = traitlets.Integer(1, config=True)
     account = traitlets.Unicode("", config=True)
     timelimit = traitlets.Unicode("", config=True)
@@ -176,8 +233,8 @@ srun -N {machines} -n {n} %s %s --profile-dir="{profile_dir}" --cluster-id="{clu
         return super(BcbioSLURMEngineSetLauncher, self).start(n)
 
 
-class BcbioSLURMControllerLauncher(SLURMLauncher, launcher.BatchClusterAppMixin):
-    """Launch a controller using SLURM."""
+class BcbioOLDSLURMControllerLauncher(SLURMLauncher, launcher.BatchClusterAppMixin):
+    """Launch a controller using SLURM for versions < 2.6"""
     account = traitlets.Unicode("", config=True)
     timelimit = traitlets.Unicode("", config=True)
     batch_file_name = Unicode(unicode("SLURM_controller" + str(uuid.uuid4())),
@@ -204,6 +261,7 @@ class BcbioPBSEngineSetLauncher(launcher.PBSEngineSetLauncher):
     """
     batch_file_name = Unicode(unicode("pbs_engines" + str(uuid.uuid4())))
     cores = traitlets.Integer(1, config=True)
+    mem = traitlets.Unicode("", config=True)
     pename = traitlets.Unicode("", config=True)
     resources = traitlets.Unicode("", config=True)
     default_template = traitlets.Unicode("""#PBS -V
@@ -212,6 +270,7 @@ class BcbioPBSEngineSetLauncher(launcher.PBSEngineSetLauncher):
 #PBS -q {queue}
 #PBS -N bcbio-ipengine
 #PBS -t 1-{n}
+{mem}
 %s %s --profile-dir="{profile_dir}" --cluster-id="{cluster_id}"
 """% (' '.join(map(pipes.quote, launcher.ipengine_cmd_argv)),
       ' '.join(timeout_params)))
@@ -219,6 +278,10 @@ class BcbioPBSEngineSetLauncher(launcher.PBSEngineSetLauncher):
     def start(self, n):
         self.context["cores"] = self.cores
         self.context["pename"] = str(self.pename)
+        if self.mem:
+            self.context["mem"] = "#PBS -l vmem=%sgb" % self.mem
+        else:
+            self.context["mem"] = ""
         return super(BcbioPBSEngineSetLauncher, self).start(n)
 
 
@@ -255,6 +318,7 @@ class TORQUELauncher(launcher.BatchSystemLauncher):
 class BcbioTORQUEEngineSetLauncher(TORQUELauncher, launcher.BatchClusterAppMixin):
     """Launch Engines using PBS"""
     cores = traitlets.Integer(1, config=True)
+    mem = traitlets.Unicode("", config=True)
     batch_file_name = Unicode(unicode("torque_engines" + str(uuid.uuid4())),
                               config=True, help="batch file name for the engine(s) job.")
     default_template = Unicode(u"""#!/bin/sh
@@ -263,6 +327,7 @@ class BcbioTORQUEEngineSetLauncher(TORQUELauncher, launcher.BatchClusterAppMixin
 #PBS -N ipengine
 #PBS -t 1-{n}
 #PBS -l nodes=1:ppn={cores}
+{mem}
 #PBS -l walltime=239:00:00
 %s %s --profile-dir="{profile_dir}" --cluster-id="{cluster_id}"
     """ % (' '.join(map(pipes.quote, launcher.ipengine_cmd_argv)),
@@ -271,6 +336,10 @@ class BcbioTORQUEEngineSetLauncher(TORQUELauncher, launcher.BatchClusterAppMixin
     def start(self, n):
         """Start n engines by profile or profile_dir."""
         self.context["cores"] = self.cores
+        if self.mem:
+            self.context["mem"] = "#PBS -l vmem=%sgb" % self.mem
+        else:
+            self.context["mem"] = ""
         return super(BcbioTORQUEEngineSetLauncher, self).start(n)
 
 
@@ -346,8 +415,18 @@ def _start(scheduler, profile, queue, num_jobs, cores_per_job, cluster_id,
     """
     ns = "cluster_helper.cluster"
     scheduler = scheduler.upper()
+    if scheduler == "SLURM" and _slurm_is_old():
+        scheduler = "OLDSLURM"
     engine_class = "Bcbio%sEngineSetLauncher" % scheduler
     controller_class = "Bcbio%sControllerLauncher" % scheduler
+    if not (engine_class in globals() and controller_class in globals()):
+        print ("The engine and controller class %s and %s are not "
+               "defined. " % (engine_class, controller_class))
+        print ("This may be due to ipython-cluster-helper not supporting "
+               "your scheduler. If it should, please file a bug report at "
+               "http://github.com/roryk/ipython-cluster-helper. Thanks!")
+        sys.exit(1)
+
     args = launcher.ipcluster_cmd_argv + \
         ["start",
          "--daemonize=True",
@@ -358,6 +437,7 @@ def _start(scheduler, profile, queue, num_jobs, cores_per_job, cluster_id,
          "--n=%s" % num_jobs,
          "--%s.cores=%s" % (engine_class, cores_per_job),
          "--%s.resources=%s" % (engine_class, extra_params.get("resources", "")),
+         "--%s.mem='%s'" % (engine_class, extra_params.get("mem", "")),
          "--IPClusterStart.controller_launcher_class=%s.%s" % (ns, controller_class),
          "--IPClusterStart.engine_launcher_class=%s.%s" % (ns, engine_class),
          "--%sLauncher.queue='%s'" % (scheduler, queue),
@@ -367,20 +447,21 @@ def _start(scheduler, profile, queue, num_jobs, cores_per_job, cluster_id,
     if scheduler in ["SGE"]:
         args += ["--%s.pename=%s" % (engine_class, _find_parallel_environment())]
     elif scheduler in ["SLURM"]:
-        # SLURM cannot get resource atts (native specification) as SGE does
-        slurm_atrs = {}
-        extra_params = extra_params['resources'].split(';')
+        if _slurm_is_old():
+            # SLURM cannot get resource atts (native specification) as SGE does
+            slurm_atrs = {}
+            extra_params = extra_params['resources'].split(';')
 
-        for parm in extra_params:
-            atr = parm.split('=')
-            slurm_atrs[atr[0]] = atr[1]
-        extra_params = slurm_atrs
+            for parm in extra_params:
+                atr = parm.split('=')
+                slurm_atrs[atr[0]] = atr[1]
+            extra_params = slurm_atrs
 
-        args += ["--%s.machines=%s" % (engine_class, extra_params.get("machines", "1"))]
-        args += ["--%s.account=%s" % (engine_class, extra_params["account"])]
-        args += ["--%s.account=%s" % (controller_class, extra_params["account"])]
-        args += ["--%s.timelimit=%s" % (engine_class, extra_params["timelimit"])]
-        args += ["--%s.timelimit=%s" % (controller_class, extra_params["timelimit"])]
+            args += ["--%s.machines=%s" % (engine_class, extra_params.get("machines", "1"))]
+            args += ["--%s.account=%s" % (engine_class, extra_params["account"])]
+            args += ["--%s.account=%s" % (controller_class, extra_params["account"])]
+            args += ["--%s.timelimit=%s" % (engine_class, extra_params["timelimit"])]
+            args += ["--%s.timelimit=%s" % (controller_class, extra_params["timelimit"])]
 
     subprocess.check_call(args)
     return cluster_id
@@ -434,7 +515,7 @@ def cluster_view(scheduler, queue, num_jobs, cores_per_job=1, profile=None,
 
     cluster_id = str(uuid.uuid4())
     url_file = get_url_file(profile, cluster_id)
-    #cluster_id = ""
+
     while 1:
         try:
             _start(scheduler, profile, queue, num_jobs, cores_per_job, cluster_id, extra_params)
@@ -469,6 +550,13 @@ def _get_balanced_blocked_view(client, retries):
         view.set_flags(retries=int(retries))
     return view
 
+def _slurm_is_old():
+    return LooseVersion(_slurm_version()) < LooseVersion("2.6")
+
+def _slurm_version():
+    version_line = subprocess.Popen("sinfo -V", shell=True,
+                                    stdout=subprocess.PIPE).communicate()[0]
+    return version_line.split()[1]
 
 # ## Temporary profile management
 
